// atoms
// classes that help find combinations of expressions
// mouser@donationcoder.com

package atompkg

import "strconv"
import "fmt"
import "sort"
import "strings"


//---------------------------------------------------------------------------
// see also http://lisperator.net/blog/find-all-expressions-that-evaluate-to-some-value/
// see also http://www.sysexpand.com/?path=exercises/expression-from-numbers
//---------------------------------------------------------------------------









//---------------------------------------------------------------------------
// The EquationAtom structure holds a string expression like "2" or "(2+3)*10" and its evaluated(cached) value

type EquationAtom struct {
	expression string
	value      int
}


// constructor to make an atom from an integer
func MakeAtomFromInteger(intval int) *EquationAtom {
	eqatom := new(EquationAtom)
	eqatom.expression = strconv.Itoa(intval)
	eqatom.value = intval
	return eqatom
}


// constructor to make an atom from an expression and val
func MakeAtomFromExpressionVal(expression string, val int) *EquationAtom {
	eqatom := new(EquationAtom)
	eqatom.expression = expression
	eqatom.value = val
	return eqatom
}
//---------------------------------------------------------------------------









//---------------------------------------------------------------------------
// custom sorting of expression array
// this is essential in order for us to canonicalize expressions, and makes it so we never see something like "-2 + 1" and instead always "1 - 2"
// see https://gobyexample.com/sorting-by-functions
type CustomExpressionStringSortableList []string


func (s CustomExpressionStringSortableList) Len() int {
	return len(s)
}


func (s CustomExpressionStringSortableList) Swap(i, j int) {
	s[i], s[j] = s[j], s[i]
}


func (s CustomExpressionStringSortableList) Less(i, j int) bool {
	var ci = s[i][0]
	var cj = s[j][0]
	// negatives always follow non-negatives
	if ci == '-' && cj != '-' {
		return false
	} else if cj == '-' && ci != '-' {
		return true
	}
	// parens always last
	if ci == '(' && cj != '(' {
		return false
	} else if cj == '(' && ci != '(' {
		return true
	}
	// shortest first
	if len(s[i]) != len(s[j]) {
		return len(s[i]) < len(s[j])
	}
	// lastly lexigraphically
	return s[i] < s[j]
}
//---------------------------------------------------------------------------




















//---------------------------------------------------------------------------
// code for combining and canonicalizing expressions

// combine two atoms with an operator
func CombinedAtomExpression(atomleft EquationAtom, atomright EquationAtom, opstring string) string {
	var expleft = atomleft.expression
	var expright = atomright.expression
	//fmt.Println("DEBUG in with " + expleft + opstring + expright + "")
	if false {
		// old way, fast but can't find duplicates using canonical expressions
		return "(" + expleft + opstring + expright + ")"
	} else {
		// new CANONICAL way - given any expression we canonicalize/normalize it so that we can find equivelent duplicates
		// first build string LISTS of expressions on left and right that can combine simply with opstring
		var expleftlist = MakeOpCompatibleListFromExpression(expleft, opstring)
		var exprightlist = MakeOpCompatibleListFromExpression(expright, opstring)
		if opstring == "-" {
			// subtraction is handled specially -- we want to treat it like addition, but negating right hand side
			opstring = "+"
			NegateAtomExpressionList(exprightlist)
		}

		// now combine left and right lists into one list we will recombine with opstring
		var allitems []string
		allitems = append(allitems, expleftlist...)
		allitems = append(allitems, exprightlist...)
		// now build new expression combining them in canonical order
		var retstr = CanonicalCombineOpList(opstring, allitems)
		//fmt.Println("DEBUG OUT with (" + retstr + ")")
		return retstr
	}
}


// make a list of strings in a longer expression that are compatible with being joined by an operation
func MakeOpCompatibleListFromExpression(expression string, opstring string) []string {
	var explist []string
	// walk expression building chunks combined with opstring
	// in other words, we only care if the NEW combining opstring is compatible with existing ops in this expression; e.g. "2+3" is compatible with a +
	// if we find chunks combined with anything else, just wrap whole thing in () and return single item
	//fmt.Println("DEBUG in MakeOpCompatibleListFromExpression with '"+expression+"' and opstring is "+opstring)
	var parendepth = 0
	var leftpos = 0
	var iscompat = false
	var lastword = ""
	// add ; at end of expression so we can more easily detect end character in range expression
	var eexpression = expression
	eexpression += ";"
	for i, c := range eexpression {
		// we dont look in parentheses
		if c == '(' {
			parendepth++
		} else if c == ')' {
			parendepth--
			// skip over digits, we will extract this word we are passing over later
		} else if c >= '0' && c <= '9' {
			// digit, just skip it
			// ok a non-digit when not in a paren denotes the end of a word (operation character or end character)
		} else if parendepth == 0 {
			// operation or END
			iscompat = false
			if c == ';' {
				iscompat = true
			} else if (c == '+' || c == '-') && (opstring == "+" || opstring == "-") {
				// + and - are compatible, because during extraction we treat the a-b as a + -b
				iscompat = true
			} else if c == '*' && opstring == "*" {
				iscompat = true
			}

			if iscompat {
				// add last WORD to list
				if i > leftpos {
					// if we hit a - at start of expression we will have i==leftpos==0
					lastword = eexpression[leftpos:i]
					explist = append(explist, lastword)
					//fmt.Println("got word "+lastword)
				}
				// advance to start of next word (leave - in the word)
				if c == '-' {
					leftpos = i
				} else {
					leftpos = i + 1
				}
			} else {
				// incompatible (e.g. we have found "2+3" and are told we want to combine this with operator *, so the only thing to do is wrap this in paren as one item
				// return whole thing wrapped in ()
				explist = make([]string, 0)
				explist = append(explist, "("+expression+")")
				break
			}
		}
	}
	return explist
}


// negate all items in the expression
func NegateAtomExpressionList(explist []string) {
	for i, exp := range explist {
		if exp[0] == '-' {
			// remove the -
			explist[i] = explist[i][1:]
		} else {
			// add a -
			explist[i] = "-" + explist[i]
		}
	}

}


// make the canonical output expression given a list of expression strings that we are combining with a single operation; e.g. "using operator + combine [1,5,-2,(15/5)]
func CanonicalCombineOpList(opstring string, itemlist []string) string {
	// sort the expression list
	if opstring == "+" || opstring == "*" {
		// these can/must be sorted for canonicalizing; we leave / along and - is never seen by us as it is converted to + before we are called
		sort.Sort(CustomExpressionStringSortableList(itemlist))
	}

	// join with op
	var retstr = strings.Join(itemlist, opstring)

	// fixups; we could do the join above in a smarter way, but because we do it automatically, we need to fix this up here
	retstr = strings.Replace(retstr, "+-", "-", -1)
	retstr = strings.Replace(retstr, "-+", "-", -1)
	retstr = strings.Replace(retstr, "--", "+", -1)

	return retstr
}
//---------------------------------------------------------------------------
























//---------------------------------------------------------------------------
// EquationTrial structures are used when we recursively take a list of atoms, choose 2 to combine, and then recursively do the same with the remainder of atoms

// the EquationTrial structure holds 2 atoms for combination via operation, and a remainder EquationSet
type EquationTrial struct {
	atom1          EquationAtom
	atom2          EquationAtom
	remainderatoms []EquationAtom
}


// the EquationTrialSet structure holds a set of EquationTrial structs
type EquationTrialSet struct {
	eqtrialslice []EquationTrial
}


// make all possible permutations of 2 atoms in the eqprimitives
func MakeEquationTrialSet(eqprimitives EquationSet) EquationTrialSet {
	// we build a trial set from the set we are passed
	var eqtrialset EquationTrialSet

	var pcount = eqprimitives.Len()
	for firsti := 0; firsti < pcount-1; firsti++ {
		for secondi := firsti + 1; secondi < pcount; secondi++ {
			var remainders []EquationAtom
			// we have chosen two items from the list (firsti and secondi)
			// so the remainders are all to left of first i plus between them, plus after secondi
			remainders = append(remainders, eqprimitives.eqslice[:firsti]...)
			remainders = append(remainders, eqprimitives.eqslice[firsti+1:secondi]...)
			remainders = append(remainders, eqprimitives.eqslice[secondi+1:]...)
			// add a new trial set from the tuple of first,secondi, remainders
			eqtrialset.MakeAppendTrial(eqprimitives.eqslice[firsti], eqprimitives.eqslice[secondi], remainders)
		}
	}
	return eqtrialset
}


func (eqtrialset *EquationTrialSet) MakeAppendTrial(atom1 EquationAtom, atom2 EquationAtom, remainderslice []EquationAtom) {
	var eqtrial EquationTrial
	eqtrial.atom1 = atom1
	eqtrial.atom2 = atom2
	eqtrial.remainderatoms = remainderslice
	//	fmt.Print("Debug MakeAppendTrial with atom1 = '"+atom1.expression+ "' and atom2 = '"+atom2.expression+"' and remainder is:")
	//	fmt.Println(remainderslice)
	// append this trial to our list
	eqtrialset.eqtrialslice = append(eqtrialset.eqtrialslice, eqtrial)
}
//---------------------------------------------------------------------------



















//---------------------------------------------------------------------------
// A slice (array) of equationatoms
type SolutionSet struct {
	// for sorting and organizing solutions by target value
	atomhash map[string]int
	solutioncount map[int]int
	maxhits_stop int
	maxhits_kill int
}


// stop after this many hits for this solutionval
func (sols *SolutionSet) Set_maxhits_stop(val int) {
	sols.maxhits_stop = val
}

// stop after this many hits for this solutionval
func (sols *SolutionSet) Set_maxhits_kill(val int) {
	sols.maxhits_kill = val
}

// Add atom to set only if unique
func (sols *SolutionSet) AddUniqueAtom(eqatom EquationAtom) {
	if sols.atomhash == nil {
		// create on first use
		sols.atomhash = make(map[string]int)
		sols.solutioncount = make(map[int]int)
	}

	// keep track of how many times we've arrived at a solution for this target value
	sols.solutioncount[eqatom.value]++

	// too many hits, remove all traces of it	
	if sols.solutioncount[eqatom.value] == sols.maxhits_kill {
		// too many hits for this target value, don't save it; this can help us save memory
		// so remove ALL solutions with this target value
		for expr, value := range sols.atomhash {
			if (value==eqatom.value) {
				delete (sols.atomhash,expr)
				}
			}
		return
	}

	// if we already stored this solution, its not a unique solution, so just return
	_, wasfound := sols.atomhash[eqatom.expression]
	if wasfound {
		// already stored this exact expression, just return and do nothing
		return
	}

	if sols.solutioncount[eqatom.value] > sols.maxhits_stop {
		// too many hits for this target value, don't save it; this can help us save memory
		return
	}
	// store value in hash of expression
	sols.atomhash[eqatom.expression] = eqatom.value
}


func (sols *SolutionSet) Len_stored() int {
	return len(sols.atomhash)
}

func (sols *SolutionSet) Len_found() int {
	var count = 0
	for _, val := range sols.solutioncount {
		count += val
		}
	return count
}
//---------------------------------------------------------------------------














//---------------------------------------------------------------------------
// for sorting and organizing solutions by target value
type AtomValueMap map[int][]string

// build a hash indexed by the target value, where each element is the list of unique expressions that reach that target
func (eqarray *EquationSet) BuildReverseValueMap() AtomValueMap {
	// build answer map
	var valuemap AtomValueMap
	valuemap = make(map[int][]string)

	// iterate all eqations
	for _, element := range eqarray.eqslice {
		valuemap[element.value] = append(valuemap[element.value], element.expression)
	}
	return valuemap
}
//---------------------------------------------------------------------------


























//---------------------------------------------------------------------------
func (sols *SolutionSet) BuildTargetValueMap() map[int][]string {
	var valuemap map[int][]string
	valuemap = make(map[int][]string)

	// iterate all eqations
	for expr, value := range sols.atomhash {
		valuemap[value] = append(valuemap[value], expr)
	}

	return valuemap
}


// dump some nice info about solutions
// solutionthresholdcount if set > 0 says not to show any target value with more than this number of solutions, as these might be too easy
// maxcharsperline lets us keep the output line length reasonable
func (sols *SolutionSet) DumpSolutionsNicely(solutionthresholdcount int, maxcharsperline int) {

	fmt.Println(" Total unique solutions found: " + strconv.Itoa(sols.Len_found()))
	fmt.Println(" Total solutions recorded: " + strconv.Itoa(sols.Len_stored()))
	
	// first lets build a reversevalue map from final values to solutions
	var valuemap = sols.BuildTargetValueMap()

	// now build an array of the keys (answers) in sorted order
	var keys []int
	for k := range valuemap {
		keys = append(keys, k)
	}
	sort.Ints(keys)

	//	var solutionword = ""
	for _, k := range keys {
		// now many solutions to this target key?
		var solcount = sols.solutioncount[k]
		var lenk = len(valuemap[k])
		if solutionthresholdcount <= 0 || solcount <= solutionthresholdcount {
			if (lenk < solcount ) {
				fmt.Print("  Answer " + strconv.Itoa(k) + " has " + strconv.Itoa(solcount) + " [" + strconv.Itoa(lenk) + "] solution(s): ")
			} else {
				fmt.Print("  Answer " + strconv.Itoa(k) + " has " + strconv.Itoa(lenk) + " solutions(s): ")
			}

			// sort to solutions to this target? (simple lexographic sort, not much reason to do this)
			// sort.Strings(valuemap[k])

			// ok now display the solutions, but stop after certain number of characters to avoid being annoying
			var estr = ""
			var dottedend = false
			for _, expr := range valuemap[k] {
				if estr != "" {
					estr += " , "
				}
				if maxcharsperline > 0 && len(estr+expr) > maxcharsperline {
					estr += "..."
					dottedend = true
					break
				}
				estr += expr
			}
			if (!dottedend && lenk < solcount) {
				estr += ", ..."
			}
			fmt.Println(estr)
		}
	}
}
//---------------------------------------------------------------------------
























//---------------------------------------------------------------------------
// A slice (array) of equationatoms
type EquationSet struct {
	eqslice       []EquationAtom
	expressionmap map[string]int
}


// Add to EquationSet from primitive integers
func (eqarray *EquationSet) AddFromIntegerSlice(intslice []int) {
	for _, element := range intslice {
		eqatom := MakeAtomFromInteger(element)
		eqarray.AddAtom(*eqatom)
	}
}


// Add to EquationSet from an atom
func (eqarray *EquationSet) AddAtom(eqatom EquationAtom) {
	eqarray.eqslice = append(eqarray.eqslice, eqatom)
}


// Add to EquationSet from slice
func (eqarray *EquationSet) AddAtomSlice(atomslice []EquationAtom) {
	eqarray.eqslice = append(eqarray.eqslice, atomslice...)
}


// number of atom elements in EquationSet
func (eqarray *EquationSet) Len() int {
	return len(eqarray.eqslice)
}
//---------------------------------------------------------------------------
























//---------------------------------------------------------------------------
// solver code

// here we have a bunch of functions that add to the solution set
// they recursively call one another as they explore the tree of possible ways to continuously take a list of expressions, combine two, and continue down the tree
// when we hit a leaf that has combined every primitive into a single expression, we add it to the solution set

// we are passed a set of primitives and want to solve and ADD to eqsolutions
// this is the (only) function that eventually gets called at the leaf level when there is only one leaf atom in our expression and it can be considered a solution, so it all comes down to this eventually
func (eqsolutions *SolutionSet) SolveFromEquationSet(eqprimitives EquationSet) {
	//
	var primitivecount = eqprimitives.Len()

	// if eqprimitives has 0 items, we just return
	if primitivecount == 0 {
		return
	}
	// if eqprimitives has 1 item, we add it as a solution
	if primitivecount == 1 {
		eqsolutions.AddUniqueAtom(eqprimitives.eqslice[0])
		//		fmt.Print("DEBUG ADDED SOLUTION: ")
		//		fmt.Println(eqprimitives.eqslice[0])
		return
	}

	// ok we need to combine primitive items
	// we do this by iterating through every (n choose 2) binomail pair of items
	// and then for each pair we will iterate through every legal combination of operators on them, and solve for a new set of primitives

	// make a trial set, which is a set of items, each of which contains 2 chosen elements, and then the remainder as another set
	var trialset = MakeEquationTrialSet(eqprimitives)

	// now solve given the trial set (add solutions to eqsolutions)
	eqsolutions.SolveFromTrialSet(trialset)
}


// we are passed a trial set, which is a set of EquationTrials, just loop over them and solve for each, adding solutiosn to our list of solutions
func (eqsolutions *SolutionSet) SolveFromTrialSet(trialset EquationTrialSet) {
	// loop each trial and solve for it
	for _, element := range trialset.eqtrialslice {
		eqsolutions.SolveFromTrial(element)
	}
}


// we are passed a trial set, which is 2 atom elements to be combined via an operator, and then a remainder set of atoms
// for each legal way to combine the 2 atoms using operators, we will combine them to generate a new atom, generate a new equationset with that new atom and the remainder
// and call SolveFromEquationSet() on that new equationset
func (eqsolutions *SolutionSet) SolveFromTrial(trial EquationTrial) {
	// all the ways to combine atom1 with atom2

	var combinedexpression string
	var combinedvalue int
	//
	var atom1val = trial.atom1.value
	var atom2val = trial.atom2.value
	//
	// given any two items, create a new expression using them, and EVERY legal operation between them
	//
	// + add them; order not important, always legal (the order will be adjusted at canonicalization step later)
	combinedexpression = CombinedAtomExpression(trial.atom1, trial.atom2, "+")
	combinedvalue = atom1val + atom2val
	eqsolutions.SolveFromNewCombo(combinedexpression, combinedvalue, trial.remainderatoms)
	//
	// * multiply them; order not important, always legal (the order will be adjusted at canonicalization step later)
	combinedexpression = CombinedAtomExpression(trial.atom1, trial.atom2, "*")
	combinedvalue = atom1val * atom2val
	eqsolutions.SolveFromNewCombo(combinedexpression, combinedvalue, trial.remainderatoms)
	//
	// - subtract them; only larger from smaller, always legal when done this way
	if atom1val >= atom2val {
		combinedexpression = CombinedAtomExpression(trial.atom1, trial.atom2, "-")
		combinedvalue = atom1val - atom2val
	} else {
		combinedexpression = CombinedAtomExpression(trial.atom2, trial.atom1, "-")
		combinedvalue = atom2val - atom1val
	}
	eqsolutions.SolveFromNewCombo(combinedexpression, combinedvalue, trial.remainderatoms)
	//
	// / divide them; only larger by smaller, only legal if no remainder and smaller is !=0
	if atom1val >= atom2val {
		if atom2val != 0 && atom1val%atom2val == 0 {
			combinedexpression = CombinedAtomExpression(trial.atom1, trial.atom2, "/")
			combinedvalue = atom1val / atom2val
			eqsolutions.SolveFromNewCombo(combinedexpression, combinedvalue, trial.remainderatoms)
		}
	} else {
		if atom1val != 0 && atom2val%atom1val == 0 {
			combinedexpression = CombinedAtomExpression(trial.atom2, trial.atom1, "/")
			combinedvalue = atom2val / atom1val
			eqsolutions.SolveFromNewCombo(combinedexpression, combinedvalue, trial.remainderatoms)
		}
	}
}


// we are passed a new combined expression and its value, and the remainder atoms
// we combine them to generate a new atom, generate a new equationset with that new atom and the remainder
// and call SolveFromEquationSet() on that new equationset
func (eqsolutions *SolutionSet) SolveFromNewCombo(combinedexpression string, combinedvalue int, remainderatoms []EquationAtom) {
	// first make new atom
	var newatom = MakeAtomFromExpressionVal(combinedexpression, combinedvalue)
	// now make new EquationSet from this atom and the remainders
	var neweqset EquationSet
	neweqset.AddAtomSlice(remainderatoms)
	neweqset.AddAtom(*newatom)
	// now solve on the new equation set
	eqsolutions.SolveFromEquationSet(neweqset)
}
//---------------------------------------------------------------------------



